# PEP0440 compatible formatted version, see:
# https://www.python.org/dev/peps/pep-0440/
#
# Generic release markers:
#   X.Y.0   # For first release after an increment in Y
#   X.Y.Z   # For bugfix releases
__version__ = '0.2.0'

import logging
from asyncio import CancelledError, create_task, iscoroutinefunction
from typing import Any, Awaitable, Callable, Dict, List, Tuple, Type, TypeVar, Union

from .event import Event

EventType = TypeVar('EventType', bound=Event)
CallableType = Callable[[EventType, Any], None]
AsyncCallableType = Callable[[EventType, Any], Awaitable[None]]
ListenerType = Union[CallableType, AsyncCallableType]
ListenersType = Dict[
    Type[Event],
    List[
        Tuple[
            Union[CallableType, AsyncCallableType],
            Tuple[Tuple[Any, ...], Dict[str, Any]],
        ]
    ]
]

_listeners: ListenersType = {}


class ListenerAlreadyExistsError(ValueError):
    pass


def listen(event_type: Type[Event], listener: ListenerType, check_if_exists: bool = False, *args, **kwargs) -> None:
    logging.debug(f'{listener} listening {event_type}')

    listeners = _listeners.get(event_type) or []

    if check_if_exists:
        for l_ in listeners:
            if l_ != listener:
                continue

            raise ListenerAlreadyExistsError

    listeners.append((listener, (args, kwargs)))

    _listeners.update({event_type: listeners})


def unregister(event_type: Type[Event], listener: ListenerType) -> None:
    if event_type not in _listeners:
        return

    _listeners[event_type] = list(filter(lambda x: x[0] != listener, _listeners[event_type]))


def unregister_all() -> None:
    global _listeners

    _listeners = {}


def on(event_type: Type[Event], *args, **kwargs):
    def register(listener: ListenerType) -> Callable[[Event], Awaitable[None]]:
        listen(event_type, listener, *args, **kwargs)

        return listener

    return register


def emit(e: Event) -> None:
    create_task(__run_listeners(e))


def get_listeners(
        event_type: Type[Event]
) -> List[Tuple[Union[CallableType, AsyncCallableType], Tuple[Tuple[Any, ...], Dict[str, Any]]]]:
    return _listeners.get(event_type, [])


async def __run_listeners(e: Event) -> None:
    listeners = _listeners.get(e.__class__) or []

    for (listener, (args, kwargs)) in listeners:
        create_task(__run_listener(e, listener, *args, **kwargs))


async def __run_listener(e: Event, listener: ListenerType, *args, **kwargs) -> None:
    # noinspection PyBroadException
    try:
        if iscoroutinefunction(listener):
            await listener(e, *args, **kwargs)

        else:
            listener(e, *args, **kwargs)

    except CancelledError:
        pass

    except Exception:
        logging.exception(f'Exception in {listener.__name__} for {e.__class__}!')



from setuptools import setup


def get_version():
    with open("events_manager/__init__.py") as f:
        for line in f.read().splitlines():
            if line.startswith('__version__'):
                return line.split('"' if '"' in line else "'")[1]

        else:
            raise RuntimeError("Unable to find version string.")


def get_requirements():
    with open("requirements.txt") as f:
        return [
            line.split('#', 1)[0].strip() for line in f.read().splitlines()
            if not line.strip().startswith('#')
        ]


def get_long_description():
    with open("README.md", encoding="utf-8") as f:
        return f.read()


setup(
    name="events-manager",
    version=get_version(),
    author="webfucktory",
    author_email="root@webfucktory.com",
    description="Python implementation for the Events Management system (aka Observer pattern)",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/webfucktory/python-events-manager",
    packages=['events_manager'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
    ],
    python_requires='>=3.8',
    install_requires=get_requirements(),
)
